//----------------------------------------------------------------------------
// F␳tB^
// Copyright (c) 2006 N099
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
// ^[Qbg̃eNX`
//----------------------------------------------------------------------------
texture txSrc;

sampler2D samSrc =
sampler_state
{
    Texture = <txSrc>;
    AddressU = Clamp;
    AddressV = Clamp;
    MinFilter = Point;
    MagFilter = Point;
};

//----------------------------------------------------------------------------
// tB^W
//----------------------------------------------------------------------------
float4 ycOffs;   //YCItZbg
float4 ycAmp;    //YC{
float4 bgrOffs;  //RGBItZbg
float4 bgrAmp;   //RGB{
float4 bgrGamma; //RGBK}

//----------------------------------------------------------------------------
// BGR <-> YC
//----------------------------------------------------------------------------
float4x4 matYC2BGR;
float4x4 matBGR2YC;
  
//-----------------------------------------------------------------------------
// Pixel Shader: ColorCorrectionPS
// Desc: F␳
//-----------------------------------------------------------------------------
float4 ColorCorrectionPS(float2 Tex : TEXCOORD0) : COLOR0
{
	float4 tmp = tex2D(samSrc, Tex) * ycAmp + ycOffs;
	float4 bgr = mul(matYC2BGR, tmp);
	tmp = saturate(bgr);
	bgr -= tmp;
	bgr += pow(tmp, bgrGamma);
	bgr = (bgr * bgrAmp) + bgrOffs;
	return mul(matBGR2YC, bgr);
}

//-----------------------------------------------------------------------------
// Pixel Shader: ColorCorrectionU16PS
// Desc: F␳iunsigned shortƈĂƑzj
//-----------------------------------------------------------------------------
float4 ColorCorrectionU16PS(float2 Tex : TEXCOORD0) : COLOR0
{
	float4 bgr,mask,tmp = tex2D(samSrc, Tex) * 65535.0f;
	
	//[0.0..1.0] -> [0..65535] -> [-32768..32767] -> 4096==1.0ɐK
	mask = step(float4(32768.0f, 32768.0f, 32768.0f, 32768.0f), tmp) * 65536.0f;
	tmp = (tmp - mask) / 4096.0f;
	
	tmp = tmp * ycAmp + ycOffs;
	bgr = mul(matYC2BGR, tmp);
	tmp = saturate(bgr);
	bgr -= tmp;
	bgr += pow(tmp, bgrGamma);
	bgr = (bgr * bgrAmp) + bgrOffs;
	tmp = mul(matBGR2YC, bgr);
	
	//͈͂tϊ
	tmp *= 4096.0f;
	tmp = clamp(tmp, -32768.0f, 32767.0f);
	mask = step(float4(0.0f, 0.0f, 0.0f, 0.0f), tmp);
	tmp += (float4(1.0f, 1.0f, 1.0f, 1.0f) - mask) * 65536.0f;
	tmp /= 65535.0f;
	
	return  tmp;
}

//-----------------------------------------------------------------------------
// Pixel Shader: ColorCorrectionS16PS
// Desc: F␳ishortƈĂƑzj
//-----------------------------------------------------------------------------
float4 ColorCorrectionS16PS(float2 Tex : TEXCOORD0) : COLOR0
{
	float4 bgr,mask,tmp = tex2D(samSrc, Tex);
	
	tmp = (tmp * 32767.0f) / 4096.0f;
	
	tmp = tmp * ycAmp + ycOffs;
	bgr = mul(matYC2BGR, tmp);
	tmp = saturate(bgr);
	bgr -= tmp;
	bgr += pow(tmp, bgrGamma);
	bgr = (bgr * bgrAmp) + bgrOffs;
	tmp = mul(matBGR2YC, bgr);
	
	tmp = (tmp * 4096.0f) / 32767.0f;
	
	return  tmp;
}

//-----------------------------------------------------------------------------
// Pixel Shader: ColorCorrectionU10PS
// Desc: F␳iin:unsigned short out:unsigned 10bitj
//-----------------------------------------------------------------------------
float4 ColorCorrectionU10PS(float2 Tex : TEXCOORD0) : COLOR0
{
	float4 bgr,mask,tmp = tex2D(samSrc, Tex) * 65535.0f;
	
	//[0.0..1.0] -> [0..65535] -> [-32768..32767] -> 4096==1.0ɐK
	mask = step(float4(32768.0f, 32768.0f, 32768.0f, 32768.0f), tmp) * 65536.0f;
	tmp = (tmp - mask) / 4096.0f;
	
	tmp = tmp * ycAmp + ycOffs;
	bgr = mul(matYC2BGR, tmp);
	tmp = saturate(bgr);
	bgr -= tmp;
	bgr += pow(tmp, bgrGamma);
	bgr = (bgr * bgrAmp) + bgrOffs;
	tmp = mul(matBGR2YC, bgr);
	
	//͈͂tϊ
	const float4 min10 = float4(0.0f,     0.0f, -512.0f, -512.0f);
	const float4 max10 = float4(0.0f,  4095.0f,  511.0f,  511.0f);
	tmp *= float4(0.0f, 2048.0f, 512.0f, 512.0f);
	
	tmp.g += 1024.0f; //Px̓ItZbg
	tmp = clamp(tmp, min10, max10);
	mask = step(0.0f, tmp);
	tmp += (1.0f - mask) * 1024.0f;
	
	//A:2bitp
	tmp.g /= 1024.0f;
	tmp.g = modf(tmp.g, tmp.r) * 1024.0f;
	tmp /= float4(3.0f, 1023.0f, 1023.0f, 1023.0f);
	
	return tmp.abgr;
}

//-----------------------------------------------------------------------------
// Technique: ColorCorrection
// Desc: F␳
//-----------------------------------------------------------------------------
technique ColorCorrection
{
	pass p0
	{
		VertexShader = null;
		PixelShader = compile ps_2_0 ColorCorrectionPS();
		ZEnable = false;
	}
}

//-----------------------------------------------------------------------------
// Technique: ColorCorrectionU16
// Desc: F␳
//-----------------------------------------------------------------------------
technique ColorCorrectionU16
{
	pass p0
	{
		VertexShader = null;
		PixelShader = compile ps_2_0 ColorCorrectionU16PS();
		ZEnable = false;
	}
}

//-----------------------------------------------------------------------------
// Technique: ColorCorrectionS16
// Desc: F␳
//-----------------------------------------------------------------------------
technique ColorCorrectionS16
{
	pass p0
	{
		VertexShader = null;
		PixelShader = compile ps_2_0 ColorCorrectionS16PS();
		ZEnable = false;
	}
}

//-----------------------------------------------------------------------------
// Technique: ColorCorrectionU10
// Desc: F␳
//-----------------------------------------------------------------------------
technique ColorCorrectionU10
{
	pass p0
	{
		VertexShader = null;
		PixelShader = compile ps_2_0 ColorCorrectionU10PS();
		ZEnable = false;
	}
}
