//----------------------------------------------------------------------------------
// aviutl  GPUF␳tB^
// Copyright (c) 2006,2007  N099
//----------------------------------------------------------------------------------
#include <windows.h>

#include <math.h>
#include <float.h>
#include <malloc.h>
#include <mmintrin.h>
#include <xmmintrin.h>
#include <new>

#include <d3d9.h>
#include <d3dx9.h>
#include <d3dx9math.h>
#include <dxerr.h>

#include "resource.h"
#include "filter.h"

#include "gpufilter.h"

//---------------------------------------------------------------------
// internal function prototypes
//---------------------------------------------------------------------
static bool searchD3D(FILTER*, IDirect3D9*&, IDirect3DDevice9*&);
static bool isSSEavailable(void);
static void yc_f16(void*, const PIXEL_YC*, void*, int, int, int, int, int);
static void f16_yc(PIXEL_YC*, const void*, void*, int, int, int, int, int);
static void yc_f32(void*, const PIXEL_YC*, void*, int, int, int, int, int);
static void f32_yc(PIXEL_YC*, const void*, void*, int, int, int, int, int);
static void yc_f16_sse(void*, const PIXEL_YC*, void*, int, int, int, int, int);
static void f16_yc_sse(PIXEL_YC*, const void*, void*, int, int, int, int, int);
static void yc_f32_sse(void*, const PIXEL_YC*, void*, int, int, int, int, int);
static void f32_yc_sse(PIXEL_YC*, const void*, void*, int, int, int, int, int);
static void yc_i16_mmx(void*, const PIXEL_YC*, void*, int, int, int, int, int);
static void i16_yc_mmx(PIXEL_YC*, const void*, void*, int, int, int, int, int);
static void i10_yc_mmx(PIXEL_YC*, const void*, void*, int, int, int, int, int);

//---------------------------------------------------------------------
// struct
//---------------------------------------------------------------------
struct PPVERT
{
    float x, y, z, rhw;
    float tu, tv;       // Texcoord

    const static D3DVERTEXELEMENT9 Decl[3];
};

const D3DVERTEXELEMENT9 PPVERT::Decl[3] =
{
	{ 0, 0,  D3DDECLTYPE_FLOAT4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITIONT, 0 },
	{ 0, 16, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,  0 },
	D3DDECL_END()
};

//---------------------------------------------------------------------
// public functions
//---------------------------------------------------------------------

//RXgN^
GpuFilter::GpuFilter() :
	m_pD3D(NULL), m_pd3dDevice(NULL),
	m_pTexSrc(NULL), m_pSrfDst(NULL), m_pSrfSys(NULL),
	m_pVert(NULL), m_pVertDec(NULL),
	m_pEffect(NULL), m_pQuery(NULL),
	m_pFloatTmp(NULL)
{
	m_width = m_height = m_wMax = -1;

	m_pixFormat = Unknown;
	ZeroMemory(m_pixFormatAvailable, sizeof(m_pixFormatAvailable));

	ZeroMemory(m_hParam, sizeof(m_hParam));
	ZeroMemory(m_hMat, sizeof(m_hMat));
	ZeroMemory(m_hTec, sizeof(m_hTec));

	m_sse = isSSEavailable();
}

//fXgN^
GpuFilter::~GpuFilter()
{
	//\[X̊J
	_aligned_free(m_pFloatTmp);
	SAFE_RELEASE(m_pQuery);
	SAFE_RELEASE(m_pEffect);
	SAFE_RELEASE(m_pVertDec);
	SAFE_RELEASE(m_pVert);
	SAFE_RELEASE(m_pSrfSys);
	SAFE_RELEASE(m_pSrfDst);
	SAFE_RELEASE(m_pTexSrc);

	//AddRefĂ
	SAFE_RELEASE(m_pd3dDevice);
	SAFE_RELEASE(m_pD3D);
}

//IuWFNg̏
bool GpuFilter::Init(FILTER *fp)
{
	HRESULT hr;

	//Ŝ
	SAFE_RELEASE(m_pd3dDevice);
	SAFE_RELEASE(m_pD3D);

	//ꉞ̃vOCŃfoCX쐬ĂȂׂ
	if(true /*!searchD3D(fp, m_pD3D, m_pd3dDevice)*/){
		//init D3D    
		m_pD3D = Direct3DCreate9(D3D_SDK_VERSION);
		if(NULL == m_pD3D) {
			ShowMessageBox(__FILE__, (DWORD)__LINE__, E_FAIL, TEXT("Direct3DCreate9()s܂"));
			return false;
		}

		//create device
		D3DPRESENT_PARAMETERS d3dpp; 

		ZeroMemory(&d3dpp, sizeof(d3dpp));
		d3dpp.Windowed = TRUE;
		d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
		d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;

		hr = m_pD3D->CreateDevice(
			D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, fp->hwnd,
			D3DCREATE_SOFTWARE_VERTEXPROCESSING|D3DCREATE_NOWINDOWCHANGES, 
			&d3dpp, &m_pd3dDevice );
		if(FAILED(hr)) {
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DfoCX̍쐬Ɏs܂"));
			SAFE_RELEASE(m_pd3dDevice);
			SAFE_RELEASE(m_pD3D);
			return false;
		}
	}

	//foCX̃`FbN
	if(FAILED(CheckDeviceCap())){
		SAFE_RELEASE(m_pd3dDevice);
		SAFE_RELEASE(m_pD3D);
		return false;
	}


	// ɑ݂郊\[X쐬

	//GtFNg̃[h
	hr = CreateEffect(MAKEINTRESOURCE(IDR_EFFECT_FILE), fp->dll_hinst);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("GtFNg̃[hɎs܂"));
		SAFE_RELEASE(m_pd3dDevice);
		SAFE_RELEASE(m_pD3D);
		return false;
	}

	//_錾̍쐬
	SAFE_RELEASE(m_pVertDec);
    hr = m_pd3dDevice->CreateVertexDeclaration(PPVERT::Decl, &m_pVertDec);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("_錾̍쐬Ɏs܂"));
		SAFE_RELEASE(m_pEffect);
		SAFE_RELEASE(m_pd3dDevice);
		SAFE_RELEASE(m_pD3D);
		return false;
	}

#ifdef PROFILE_GPU
	//fobOp
    hr = CreateQuery();
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("NG̍쐬Ɏs܂"));
		SAFE_RELEASE(m_pVertDec);
		SAFE_RELEASE(m_pEffect);
		SAFE_RELEASE(m_pd3dDevice);
		SAFE_RELEASE(m_pD3D);
		return false;
	}
#endif

	return true;
}

//e탊\[X̊m
bool GpuFilter::AllocResources(int w, int h, int w_max, PIXEL_FORMAT fmt)
{
	//̃`FbN
	if(!m_pEffect) return false;

	//sNZTCỸZbgi4096x4096ɐj
	if(w<0 || w>4096 || h<0 || h>4096 || w_max<0 || w_max<w){
		return false;
	}else{
		m_width = w;
		m_height = h;
		m_wMax = w_max;
	}

	//sNZtH[}bg
	m_pixFormat = PixelFormatFromIndex(fmt);
	if(!m_pixFormatAvailable[m_pixFormat]){
		//vꂽtH[}bg͑ΉĂȂ
		m_pixFormat = Unknown;
		for(int i=0; i<Unknown; ++i){
			if(m_pixFormatAvailable[i]){
				m_pixFormat = (PIXEL_FORMAT)i;
				break;
			}
		}

		//ΉtH[}bg݂Ȃ
		if(m_pixFormat == Unknown){
			ShowMessageBox(__FILE__, (DWORD)__LINE__, E_NOTIMPL, TEXT("p\ȃtH[}bg܂"));
			return false;
		}
	}
	
	//ꉞJĂ
	ReleaseResources();

	//VXeɍ쐬郊\[X
	if(FAILED(CreateSysMemResources())){
		return false;
	}

	//GPUɍ쐬郊\[X
	if(FAILED(CreateDefaultResources())){
		return false;
	}

	//eNX`̃Zbg
	HRESULT hr = m_pEffect->SetTexture("txSrc", m_pTexSrc);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("eNX`̃ZbgɎs܂"));
		return false;
	}

	//KvȂF16pϊpe|̈쐬
	if(F16 == m_pixFormat){
		m_pFloatTmp = (__m128*)_aligned_malloc(sizeof(__m128)*m_width, sizeof(__m128));
		if(NULL == m_pFloatTmp){
			ShowMessageBox(__FILE__, (DWORD)__LINE__, E_OUTOFMEMORY, 
				TEXT("ϊp̊mۂɎs܂"));
			return false;
		}
	}

	return true;
}

//e탊\[X̊J
bool GpuFilter::ReleaseResources(void)
{	
	//\[X̊J
	_aligned_free(m_pFloatTmp);
	m_pFloatTmp = NULL;
	SAFE_RELEASE(m_pVert);
	SAFE_RELEASE(m_pSrfSys);
	SAFE_RELEASE(m_pSrfDst);
	SAFE_RELEASE(m_pTexSrc);

	return true;
}

//tB^p[^̃Zbg
bool GpuFilter::SetParams(int *track, int *check)
{
	HRESULT hr;

	//̃`FbN
	if(!m_pEffect) return false;

	//p[^ϊ
	D3DXVECTOR4 vParams;
	//Ql
	// T_YOF, T_YGA, T_CBOF, T_CBGA, T_CROF, T_CRGA,
	// T_ROF, T_RGA, T_RGM,  T_GOF,  T_GGA,  T_GGM, T_BOF, T_BGA, T_BGM

	//yc offset
	if(m_hParam[PARAM_YC_OFFS]){
		vParams.x = 0.0f;
		vParams.y = track[T_YOF]/256.0f;
		vParams.z = track[T_CBOF]/256.0f;
		vParams.w = track[T_CROF]/256.0f;
		hr = m_pEffect->SetVector(m_hParam[PARAM_YC_OFFS], &vParams);
		if(FAILED(hr)){
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, 
				TEXT("p[^̃ZbgɎs܂"));
		}
	}
	//yc gain
	if(m_hParam[PARAM_YC_GAIN]){
		vParams.x = 0.0f;
		vParams.y = (track[T_YGA]+256)/256.0f;
		vParams.z = (track[T_CBGA]+256)/256.0f;
		vParams.w = (track[T_CRGA]+256)/256.0f;
		hr = m_pEffect->SetVector(m_hParam[PARAM_YC_GAIN], &vParams);
		if(FAILED(hr)){
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, 
				TEXT("p[^̃ZbgɎs܂"));
		}
	}
	//rgb offset
	if(m_hParam[PARAM_BGR_OFFS]){
		vParams.x = 0.0f;
		vParams.w = track[T_ROF]/256.0f;
		vParams.z = track[T_GOF]/256.0f;
		vParams.y = track[T_BOF]/256.0f;
		hr = m_pEffect->SetVector(m_hParam[PARAM_BGR_OFFS], &vParams);
		if(FAILED(hr)){
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, 
				TEXT("p[^̃ZbgɎs܂"));
		}
	}
	//rgb gain
	if(m_hParam[PARAM_BGR_GAIN]){
		vParams.x = 0.0f;
		vParams.w = (track[T_RGA]+256)/256.0f;
		vParams.z = (track[T_GGA]+256)/256.0f;
		vParams.y = (track[T_BGA]+256)/256.0f;
		hr = m_pEffect->SetVector(m_hParam[PARAM_BGR_GAIN], &vParams);
		if(FAILED(hr)){
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, 
				TEXT("p[^̃ZbgɎs܂"));
		}
	}
	//rgb gamma
	if(m_hParam[PARAM_BGR_GAMMA]){
		vParams.x = 1.0f;
		vParams.w = (float)pow(8.0, track[T_RGM]/256.0);
		vParams.z = (float)pow(8.0, track[T_GGM]/256.0);
		vParams.y = (float)pow(8.0, track[T_BGM]/256.0);
		hr = m_pEffect->SetVector(m_hParam[PARAM_BGR_GAMMA], &vParams);
		if(FAILED(hr)){
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, 
				TEXT("p[^̃ZbgɎs܂"));
		}
	}

	//RGB<=>YCϊ}gNX
	//----------------------------------------------------------------------------
	// BGR <-> YC (CCIR 601-1:derived from TIFF 6.0 section 21, 1992/6/3)
	//----------------------------------------------------------------------------
	static const float mYC2BGR601[] = 
	{ 0.0f, 0.0f,      0.0f,      0.0f,
	  0.0f, 1.0f,  1.77200f,      0.0f,
	  0.0f, 1.0f, -0.34414f, -0.71414f,
	  0.0f, 1.0f,      0.0f,  1.40200f }; 

	static const float mBGR2YC601[] = 
	{ 0.0f,      0.0f,      0.0f,      0.0f,
	  0.0f,    0.114f,   0.5870f,  0.29900f,
	  0.0f,  0.50000f, -0.33126f, -0.16874f,
	  0.0f, -0.08131f, -0.41869f,  0.50000f }; 

	//----------------------------------------------------------------------------
	// BGR <-> YC (Recommendation ITU-R BT.709)
	//----------------------------------------------------------------------------
	static const float mYC2BGR709[] = 
	{ 0.0f, 0.0f,     0.0f,     0.0f,
	  0.0f, 1.0f, -1.8556f,     0.0f,
	  0.0f, 1.0f, -0.1870f, -0.4664f,
	  0.0f, 1.0f,     0.0f,  1.5701f }; 

	static const float mBGR2YC709[] = 
	{ 0.0f,      0.0f,     0.0f,     0.0f,
	  0.0f,   0.0721f,  0.7154f,  0.2125f,
	  0.0f,   0.5000f, -0.3855f, -0.1145f,
	  0.0f,  -0.0459f, -0.4556f,  0.5016f }; 

	//}gbNX̃Zbg
	bool bt709 = false;//check[C_709] ? true : false;
	hr = m_pEffect->SetValue(m_hMat[MAT_YC2BGR], 
		bt709 ? mYC2BGR709 : mYC2BGR601, sizeof(mYC2BGR709));
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, 
			TEXT("}gNX̃ZbgɎs܂"));
	}
	hr = m_pEffect->SetValue(m_hMat[MAT_BGR2YC], 
		bt709 ? mBGR2YC709 : mBGR2YC601, sizeof(mYC2BGR709));
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, 
			TEXT("}gNX̃ZbgɎs܂"));
	}

	return true;
}

//foCX̏ԂeXgAKvȂ烊Zbg
bool GpuFilter::TestDeviceState(bool *need_realloc)
{
	HRESULT hr;

	if(!m_pd3dDevice){
		return false;
	}

	// D3D9 device ̏ԂeXg
	hr = m_pd3dDevice->TestCooperativeLevel();
	switch(hr)
	{
		case D3D_OK:
			break;
		case D3DERR_DEVICELOST:
#ifdef PROFILE_GPU
				ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
					TEXT("foCXXg܂Bs܂B"));
#endif
			return false;
		case D3DERR_DEVICENOTRESET:
			hr = Reset();
			if(FAILED(hr)){
				ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
					TEXT("foCXZbgɎs܂"));
				return false;
			}else{
				*need_realloc = true;
			}
			break;
		default :
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
				TEXT("TestCooperativeLevel()G[Ԃ܂"));
			return false;
	}

	return true;
}

//PIXEL_YCf[^foCXɓ]
bool GpuFilter::PutImage(const PIXEL_YC *data)
{
	HRESULT hr;
	IDirect3DSurface9 *src = NULL;

	//eNX`T[tFX̎擾
	hr = m_pTexSrc->GetSurfaceLevel(0, &src);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("eNX`T[tFX擾ł܂ł"));
		return false;
	}
#ifdef USE_DYNAMIC_TEXTURE
	hr = PutPixelData(src, data);
#else
	hr = PutPixelData(src, m_pSrfSys, data);
#endif
	SAFE_RELEASE(src);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("foCXփRs[ł܂ł"));
		return false;
	}

	return true;
}

//PIXEL_YCf[^foCX]
bool GpuFilter::GetImage(PIXEL_YC *data)
{
	HRESULT hr;

	hr = GetPixelData(data, m_pSrfDst, m_pSrfSys);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("foCXRs[ł܂ł"));
		return false;
	}

	return true;
}

//f[^̏s
bool GpuFilter::ProcImage(void)
{
	if(!m_pSrfDst || !m_pd3dDevice){
		return false;
	}

	//_OJn
	if(SUCCEEDED(m_pd3dDevice->BeginScene()))
	{
		//KȃeNjbNw
		if(m_pixFormat == U16){
			m_pEffect->SetTechnique(m_hTec[TEC_UI16]);
		}else if(m_pixFormat == S16){
			m_pEffect->SetTechnique(m_hTec[TEC_SI16]);
		}else if(m_pixFormat == U10){
			m_pEffect->SetTechnique(m_hTec[TEC_UI10]);
		}else{
			m_pEffect->SetTechnique(m_hTec[TEC_FLOAT]);
		}

		// Set the vertex declaration
		m_pd3dDevice->SetVertexDeclaration(m_pVertDec);

		// Draw the quad
		UINT cPasses;
		m_pEffect->Begin( &cPasses, 0 );

		//RenderTarget̐ݒ
		m_pd3dDevice->SetRenderTarget(0, m_pSrfDst);

		//Render
		m_pEffect->BeginPass(0);
		m_pd3dDevice->SetStreamSource(0, m_pVert, 0, sizeof(PPVERT));
		m_pd3dDevice->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
		m_pEffect->EndPass();

		m_pEffect->End();

		//End scene
		m_pd3dDevice->EndScene();
	}

	return true;
}

//D3DC^[tF[XOɓn
bool GpuFilter::GetD3DInterface(IDirect3D9* &pD3D, IDirect3DDevice9* &pd3dDevice)
{
	if(!m_pD3D || !m_pd3dDevice){
		return false;
	}else{
		//AddRefsŃRs[
		m_pD3D->AddRef();
		m_pd3dDevice->AddRef();
		pD3D = m_pD3D;
		pd3dDevice = m_pd3dDevice; 
	}

	return true;
}

//eNX`tH[}bg̕\
const TCHAR *GpuFilter::GetPixelFormatStr(void) const
{
	switch(m_pixFormat){
		case F32:
			return TEXT("F32");
		case F16:
			return TEXT("F16");
		case U16:
			return TEXT("U16");
		case S16:
			return TEXT("S16");
		case U10:
			return TEXT("U10");
		case Unknown:
		default:
			return TEXT("Unknown");
	}
}

//GPUł̏I܂ő҂
#ifdef PROFILE_GPU
void GpuFilter::FlushCommand(void)
{
	if(m_pQuery){
		m_pQuery->Issue(D3DISSUE_END);
		while(S_FALSE == m_pQuery->GetData(NULL, 0, D3DGETDATA_FLUSH))
			__asm pause ;
	}
}
#endif

//---------------------------------------------------------------------
// private functions
//---------------------------------------------------------------------

//foCX̃T|[g󋵂̃`FbN
HRESULT GpuFilter::CheckDeviceCap(void)
{
	HRESULT hr;

	//|C^̃`FbN
	if(!m_pd3dDevice || !m_pD3D) return D3DERR_INVALIDCALL;

	//D3DCaps̃`FbN
	D3DCAPS9 caps;
	hr = m_pD3D->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &caps);
	if(D3D_OK != hr) return hr;

	//eNX`֘ÃeXg
	if(caps.TextureCaps & D3DPTEXTURECAPS_SQUAREONLY){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("񐳕eNX`𗘗pł܂"));
		return D3DERR_NOTAVAILABLE;
	}
	if(!(caps.TextureCaps & D3DPTEXTURECAPS_POW2)){
		; //2̗ݏȊÕTCỸeNX`
	}else if(!(caps.TextureCaps & D3DPTEXTURECAPS_NONPOW2CONDITIONAL)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("2̗ݏTCYȊÕeNX`͗pł܂"));
		return D3DERR_NOTAVAILABLE;
	}

    //RTtH[}bg̃`FbN
    D3DDISPLAYMODE DisplayMode;
    hr = m_pd3dDevice->GetDisplayMode( 0, &DisplayMode );
	if(D3D_OK != hr) return hr;

	bool found = false;
    if(SUCCEEDED(m_pD3D->CheckDeviceFormat(caps.AdapterOrdinal, caps.DeviceType,
		DisplayMode.Format, D3DUSAGE_RENDERTARGET, D3DRTYPE_TEXTURE, D3DFMT_A16B16G16R16F)))
	{
		found = true;
		m_pixFormatAvailable[F16] = true;
	}else{
#ifdef PROFILE_GPU
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DFMT_A16B16G16R16F pł܂"));
#endif
		m_pixFormatAvailable[F16] = false;
	}

	if(SUCCEEDED(m_pD3D->CheckDeviceFormat(caps.AdapterOrdinal, caps.DeviceType,
		DisplayMode.Format, D3DUSAGE_RENDERTARGET, D3DRTYPE_TEXTURE, D3DFMT_A32B32G32R32F)))
	{
		found = true;
		m_pixFormatAvailable[F32] = true;
	}else{
#ifdef PROFILE_GPU
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DFMT_A32B32G32R32F pł܂"));
#endif
		m_pixFormatAvailable[F32] = false;
	}

	if(SUCCEEDED(m_pD3D->CheckDeviceFormat(caps.AdapterOrdinal, caps.DeviceType,
		DisplayMode.Format, D3DUSAGE_RENDERTARGET, D3DRTYPE_TEXTURE, D3DFMT_A16B16G16R16)))
	{
		found = true;
		m_pixFormatAvailable[U16] = true;
	}else{
#ifdef PROFILE_GPU
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DFMT_A16B16G16R16 pł܂"));
#endif
		m_pixFormatAvailable[U16] = false;
	}

	if(SUCCEEDED(m_pD3D->CheckDeviceFormat(caps.AdapterOrdinal, caps.DeviceType,
		DisplayMode.Format, D3DUSAGE_RENDERTARGET, D3DRTYPE_TEXTURE, D3DFMT_Q16W16V16U16)))
	{
		found = true;
		m_pixFormatAvailable[S16] = true;
	}else{
#ifdef PROFILE_GPU
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DFMT_Q16W16V16U16 pł܂"));
#endif
		m_pixFormatAvailable[S16] = false;
	}
	
#ifdef USE_DYNAMIC_TEXTURE
	if(SUCCEEDED(m_pD3D->CheckDeviceFormat(caps.AdapterOrdinal, caps.DeviceType,
		DisplayMode.Format, D3DUSAGE_RENDERTARGET, D3DRTYPE_TEXTURE, D3DFMT_A2B10G10R10)))
	{
		found = true;
		m_pixFormatAvailable[U10] = true;
	}else{
#ifdef PROFILE_GPU
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DFMT_A2B10G10R10 pł܂"));
#endif
		m_pixFormatAvailable[U10] = false;
	}
#else
	//U10tH[}bg̓_Ci~bNeNX`p̂ݗL
	m_pixFormatAvailable[U10] = false;
#endif

	if(!found){   
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("p\ȃsNZtH[}bgЂƂ܂"));
		return D3DERR_NOTAVAILABLE;
	}

	return hr;
}

//_̐錾
HRESULT GpuFilter::CreateVert(void)
{
	HRESULT hr;

    //foCX̃`FbN
	if(!m_pd3dDevice) return D3DERR_INVALIDCALL;

	//Ŝ
	SAFE_RELEASE(m_pVert);

	const PPVERT Quad[4] =
	{
		{         -0.5f,          -0.5f, 1.0f, 1.0f, 0.0f, 0.0f },
		{ m_width -0.5f,          -0.5f, 1.0f, 1.0f, 1.0f, 0.0f },
		{         -0.5f, m_height -0.5f, 1.0f, 1.0f, 0.0f, 1.0f },
		{ m_width -0.5f, m_height -0.5f, 1.0f, 1.0f, 1.0f, 1.0f }
	};

	hr = m_pd3dDevice->CreateVertexBuffer(sizeof(PPVERT) * 4,
		D3DUSAGE_WRITEONLY, 0, D3DPOOL_DEFAULT, &m_pVert, NULL );

	//obt@ɓ]
	LPVOID pVBData;
	if(SUCCEEDED(m_pVert->Lock(0, 0, &pVBData, D3DLOCK_DISCARD|D3DLOCK_NOSYSLOCK))){
		CopyMemory(pVBData, Quad, sizeof(Quad));
		m_pVert->Unlock();
	}

	return hr;
}

//GtFNgt@CǂݍŃGtFNg쐬
HRESULT GpuFilter::CreateEffect(const TCHAR *res, HMODULE hm)
{
	const DWORD ShaderFlags = D3DXFX_NOT_CLONEABLE;
	ID3DXBuffer *pCompErr = NULL; //RpC̃G[
	HRESULT hr;

	//foCX̃`FbN
	if(!m_pd3dDevice) return D3DERR_INVALIDCALL;

	//Ŝ
	SAFE_RELEASE(m_pEffect);

	//GtFNgt@C̓ǂݎ
	hr = D3DXCreateEffectFromResource(m_pd3dDevice, hm,
		   res, NULL, NULL, ShaderFlags, NULL, &m_pEffect, &pCompErr);
	if(D3D_OK != hr){
		if(pCompErr){
			//RpCG[\
			char *msg = (char*)pCompErr->GetBufferPointer();
#ifdef UNICODE
			if(msg){
				size_t count = pCompErr->GetBufferSize();
				WCHAR *str = new(std::nothrow) WCHAR[count];
				if(str){
					mbstowcs(str, msg, count);
					str[count -1] = L'\0';
					ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, str);
					delete[] str;
				}
			}
#else
			if(msg){
				ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, msg);
			}
#endif
			pCompErr->Release();
		}
		
	}else{
		//p[^nh̎擾
		m_hParam[PARAM_YC_OFFS]   = m_pEffect->GetParameterByName(NULL, "ycOffs");
		m_hParam[PARAM_YC_GAIN]   = m_pEffect->GetParameterByName(NULL, "ycAmp");
		m_hParam[PARAM_BGR_OFFS]  = m_pEffect->GetParameterByName(NULL, "bgrOffs");
		m_hParam[PARAM_BGR_GAIN]  = m_pEffect->GetParameterByName(NULL, "bgrAmp");
		m_hParam[PARAM_BGR_GAMMA] = m_pEffect->GetParameterByName(NULL, "bgrGamma");
		m_hMat[MAT_YC2BGR] = m_pEffect->GetParameterByName(NULL, "matYC2BGR");
		m_hMat[MAT_BGR2YC] = m_pEffect->GetParameterByName(NULL, "matBGR2YC");

		//eNjbN
		m_hTec[TEC_FLOAT] = m_pEffect->GetTechniqueByName("ColorCorrection");
		m_hTec[TEC_UI16]  = m_pEffect->GetTechniqueByName("ColorCorrectionU16");
		m_hTec[TEC_SI16]  = m_pEffect->GetTechniqueByName("ColorCorrectionS16");
		m_hTec[TEC_UI10]  = m_pEffect->GetTechniqueByName("ColorCorrectionU10");
	}

	SAFE_RELEASE(pCompErr);

	return hr;
}

//VXeɕێ郊\[X쐬
HRESULT GpuFilter::CreateSysMemResources(void)
{
	HRESULT hr;

	//VXẽT[tFX
	hr = CreateSurface(&m_pSrfSys, 0, D3DPOOL_SYSTEMMEM);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("T[tFX̍쐬Ɏs܂"));
		return hr;
	}

	return hr;
}

//GPUɑ݂郊\[X쐬FXgɂ͕AKv
HRESULT GpuFilter::CreateDefaultResources(void)
{
	HRESULT hr;

	//_
	hr = CreateVert();
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("_̃ZbgɎs܂"));
		return hr;
	}

	//eNX`
#ifdef USE_DYNAMIC_TEXTURE
	hr = CreateTexture(&m_pTexSrc, D3DUSAGE_DYNAMIC, D3DPOOL_DEFAULT);
#else
	hr = CreateTexture(&m_pTexSrc, 0, D3DPOOL_DEFAULT);
#endif
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("eNX`̍쐬Ɏs܂"));
		return hr;
	}

	//_[^[Qbg
	hr = CreateSurface(&m_pSrfDst, D3DUSAGE_RENDERTARGET, D3DPOOL_DEFAULT);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("_[^[Qbg̍쐬Ɏs܂"));
		return hr;
	}

#ifdef PROFILE_GPU
	//fobOp
    hr = CreateQuery();
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("NG̍쐬Ɏs܂"));
		return hr;
	}
#endif
	return hr;
}

//foCXXgɃZbgs
HRESULT GpuFilter::Reset(void)
{
	HRESULT hr;

	//foCX̃`FbN
	if(!m_pd3dDevice || !m_pEffect) return D3DERR_INVALIDCALL;

	//GPU\[X̊J
	SAFE_RELEASE(m_pVert);
	SAFE_RELEASE(m_pSrfDst);
	SAFE_RELEASE(m_pTexSrc);
#ifdef PROFILE_GPU
	SAFE_RELEASE(m_pQuery);
#endif
	m_pEffect->OnLostDevice();

	//Zbg
	D3DPRESENT_PARAMETERS dpp;

	ZeroMemory(&dpp, sizeof(dpp));
	dpp.Windowed = TRUE;
	dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	dpp.BackBufferFormat = D3DFMT_UNKNOWN;

	hr = m_pd3dDevice->Reset(&dpp);
	if(FAILED(hr)){
		return hr;
	}

	hr = m_pEffect->OnResetDevice();
	if(FAILED(hr)){
		return hr;
	}

	//\[X̍Ċmۂ͂ł͍sȂ
	return hr;
}

//eNX`쐬
HRESULT GpuFilter::CreateTexture(IDirect3DTexture9 **ppTex, DWORD usage, D3DPOOL pool)
{
	D3DFORMAT fmt;

	//10bittH[}bg̃\[XeNX`UI16tH[}bgƂ
	if(m_pixFormat == U10){
		fmt = PicFmtToD3DFmt(U16);
	}else{
		fmt = PicFmtToD3DFmt(m_pixFormat);
	}

	//foCX̃`FbN
	if(!m_pd3dDevice) return D3DERR_INVALIDCALL;
	
	//Ŝ
	SAFE_RELEASE(*ppTex);

	//eNX`쐬
	return m_pd3dDevice->CreateTexture(m_width, m_height, 1, usage, fmt, pool, ppTex, NULL);
}

//T[tFX쐬
HRESULT GpuFilter::CreateSurface(IDirect3DSurface9 **ppSrf, DWORD usage, D3DPOOL pool)
{
	D3DFORMAT fmt = PicFmtToD3DFmt(m_pixFormat);

	//foCX̃`FbN
	if(!m_pd3dDevice) return D3DERR_INVALIDCALL;

	//Ŝ
	SAFE_RELEASE(*ppSrf);

	if(usage & D3DUSAGE_RENDERTARGET){
		//_[^[Qbg쐬
		return m_pd3dDevice->CreateRenderTarget(m_width, m_height, fmt, D3DMULTISAMPLE_NONE, 0, FALSE, ppSrf, NULL);
	}else{
		//ItXN[T[tFX쐬
		return m_pd3dDevice->CreateOffscreenPlainSurface(m_width, m_height, fmt, pool, ppSrf, NULL);
	}
}

//UpdateTexture𗘗pԐړ]
HRESULT GpuFilter::PutPixelData(IDirect3DSurface9 *dst, IDirect3DSurface9 *sys, const PIXEL_YC *src)
{
	HRESULT hr;
	D3DLOCKED_RECT lRect;

	if(!m_pd3dDevice || !sys) return D3DERR_INVALIDCALL;

	//VXẽT[tFXbN
	hr = sys->LockRect(&lRect, NULL, D3DLOCK_DISCARD|D3DLOCK_NOSYSLOCK);
	if(hr != D3D_OK) return hr;

	//f[^̃Rs[
	if(U16 == m_pixFormat || S16 == m_pixFormat || U10 == m_pixFormat){
		yc_i16_mmx(lRect.pBits, src, NULL, 
			m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m64), m_height);
	}else if(F32 == m_pixFormat){
		if(m_sse){
			yc_f32_sse(lRect.pBits, src, NULL, 
				m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m128), m_height);
		}else{
			yc_f32(lRect.pBits, src, NULL, 
				m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m128), m_height);
		}
	}else{
		if(m_sse){
			yc_f16_sse(lRect.pBits, src, m_pFloatTmp, 
				m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m64), m_height);
		}else{
			yc_f16(lRect.pBits, src, m_pFloatTmp, 
				m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m64), m_height);
		}
	}
	sys->UnlockRect();

	//f[^̓]
	return m_pd3dDevice->UpdateSurface(sys, NULL, dst, NULL);
}

//DynamicTexture𗘗pړ]
HRESULT GpuFilter::PutPixelData(IDirect3DSurface9 *dst, const PIXEL_YC *src)
{
	HRESULT hr;
	D3DLOCKED_RECT lRect;

	if(!dst) return D3DERR_INVALIDCALL;

	//_Ci~bNeNX`̃T[tFXbN
	hr = dst->LockRect(&lRect, NULL, D3DLOCK_DISCARD|D3DLOCK_NOSYSLOCK);
	if(hr != D3D_OK) return hr;

	//f[^̓]
	if(U16 == m_pixFormat || S16 == m_pixFormat || U10 == m_pixFormat){
		yc_i16_mmx(lRect.pBits, src, NULL, 
			m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m64), m_height);
	}else if(F32 == m_pixFormat){
		if(m_sse){
			yc_f32_sse(lRect.pBits, src, NULL, 
				m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m128), m_height);
		}else{
			yc_f32(lRect.pBits, src, NULL, 
				m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m128), m_height);
		}
	}else{
		if(m_sse){
			yc_f16_sse(lRect.pBits, src, m_pFloatTmp, 
				m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m64), m_height);
		}else{
			yc_f16(lRect.pBits, src, m_pFloatTmp, 
				m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m64), m_height);
		}
	}
	
	return dst->UnlockRect();
}

//GetRenderTargetData𗘗pԐړ]
HRESULT GpuFilter::GetPixelData(PIXEL_YC *dst, IDirect3DSurface9 *src, IDirect3DSurface9 *sys)
{
	HRESULT hr;
	D3DLOCKED_RECT lRect;

	if(!m_pd3dDevice || !sys) return D3DERR_INVALIDCALL;

	//VXeɓ]
	hr = m_pd3dDevice->GetRenderTargetData(src, sys);
	if(hr != D3D_OK) return hr;

	//VXẽT[tFXbN
	hr = sys->LockRect(&lRect, NULL, D3DLOCK_READONLY|D3DLOCK_NOSYSLOCK);
	if(hr != D3D_OK) return hr;

	//f[^̃Rs[
	if(U16 == m_pixFormat || S16 == m_pixFormat){
		i16_yc_mmx(dst, lRect.pBits, NULL, 
			m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m64), m_height);
	}else if(U10 == m_pixFormat){
		i10_yc_mmx(dst, lRect.pBits, NULL, 
			m_width, m_height, m_wMax, lRect.Pitch/sizeof(UINT), m_height);
	}else if(F32 == m_pixFormat){
		if(m_sse){
			f32_yc_sse(dst, lRect.pBits, NULL, 
				m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m128), m_height);
		}else{
			f32_yc(dst, lRect.pBits, NULL, 
				m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m128), m_height);
		}
	}else{
		if(m_sse){
			f16_yc_sse(dst, lRect.pBits, m_pFloatTmp, 
				m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m64), m_height);
		}else{
			f16_yc(dst, lRect.pBits, m_pFloatTmp, 
				m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m64), m_height);
		}
	}

	return sys->UnlockRect();
}

//G[̕\
HRESULT GpuFilter::ShowMessageBox(const char *file, DWORD line, HRESULT hr, const TCHAR *msg)
{
#ifdef _DEBUG
	return DXTrace(file, line, hr, msg, TRUE);
#else
	MessageBox(NULL, msg, TEXT("GPUF␳"), MB_ICONERROR|MB_SYSTEMMODAL);
	return S_OK;
#endif
}

//PIXEL_FORMATD3DFORMATɕϊ
D3DFORMAT GpuFilter::PicFmtToD3DFmt(PIXEL_FORMAT p) const
{
	switch(p){
		case F16:
			return D3DFMT_A16B16G16R16F;
		case U16:
			return D3DFMT_A16B16G16R16;
		case S16:
			return D3DFMT_Q16W16V16U16;
		case U10:
			return D3DFMT_A2B10G10R10;
		case F32:
		case Unknown:
		default:
			return D3DFMT_A32B32G32R32F;
	}
}

#ifdef PROFILE_GPU
//fobOpNG
HRESULT GpuFilter::CreateQuery(void)
{
	SAFE_RELEASE(m_pQuery);
	return m_pd3dDevice->CreateQuery(D3DQUERYTYPE_EVENT, &m_pQuery);
}
#endif

//---------------------------------------------------------------------
// internal functions
//---------------------------------------------------------------------
//PIXEL_YC̕ϊ
// Y (0, 4096) <-> (0.0, 1.0)
// C (-2048, 2048) <-> (-0.5, 0.5)

//t[̃TCY
// PIXEL_YC
//        w
//  ==============----- 
//  *            *    |
// h*            *    |
//  *            *    |max_h
//  *============*    |
//  |       max_w     |
//  -------------------
//
// Texture
//        w
//  ==============------- 
//  *            *      |
// h*            *      |
//  *            *      |
//  *============*      |sh
//  |                   |
//  |        sw         |
//  ---------------------
//
// float -> short ̊ۂ߂ /QIfist IvV𗘗pivs2005ł͈ӖHj 
// SSE𗘗p邱Ƃł؎̂Ăł(CVTPS2PI)

//ϊRs[ PIXEL_YC -> A16B16G16R16F
void yc_f16(void *dst, const PIXEL_YC *src, void *tmp, int w, int h, int max_w, int sw, int sh)
{
	int i,j;
	const float r = 1.0f/4096.0f;
	float *ftmp = (float*)tmp;
	D3DXFLOAT16 *stmp = (D3DXFLOAT16*)dst;

	for(i=0; i<h; ++i){
		for(j=0; j<w; ++j){
			ftmp[j*4] = 0.0f;
			ftmp[j*4 + 1] = src->y  * r;
			ftmp[j*4 + 2] = src->cb * r;
			ftmp[j*4 + 3] = src->cr * r;
			src++;
		}
		D3DXFloat32To16Array(stmp, ftmp, w*4);
		stmp += w*4;
		//ZeroMemory(stmp, (sw-w)*sizeof(D3DXFLOAT16)*4);

		src += (max_w - w);
		stmp += (sw - w)*4;
	}
	//ZeroMemory(stmp, sw*(sh-h)*sizeof(D3DXFLOAT16)*4);
}

//ϊRs[ A16B16G16R16F -> PIXEL_YC
void f16_yc(PIXEL_YC *dst, const void *src, void *tmp, int w, int h, int max_w, int sw, int sh)
{
	int i,j;
	float *ftmp = (float*)tmp;
	const D3DXFLOAT16 *stmp = (const D3DXFLOAT16*)src;
	
	for(i=0; i<h; ++i){
		D3DXFloat16To32Array(ftmp, stmp, w*4);
		for(j=0; j<w; ++j){
			dst->y  = (short)(4096.0f * ftmp[j*4 + 1]);
			dst->cb = (short)(4096.0f * ftmp[j*4 + 2]);
			dst->cr = (short)(4096.0f * ftmp[j*4 + 3]);
			dst++;
		}
		stmp += sw*4;
		dst += (max_w - w);
	}
}

//ϊRs[ PIXEL_YC -> A32B32G32R32F
void yc_f32(void *dst, const PIXEL_YC *src, void *tmp, int w, int h, int max_w, int sw, int sh)
{
	int i,j;
	const float r = 1.0f/4096.0f;
	float *fdst = (float*)dst;

	for(i=0; i<h; ++i){
		for(j=0; j<w; ++j){
			*fdst++ = 0.0f;
			*fdst++ = src->y  * r;
			*fdst++ = src->cb * r;
			*fdst++ = src->cr * r;
			src++;
		}
		//ZeroMemory(fdst, (sw-w)*sizeof(float)*4);

		src += (max_w - w);
		fdst += (sw - w)*4;
	}
	//ZeroMemory(fdst, sw*(sh-h)*sizeof(float)*4);
}

//ϊRs[ A32B32G32R32F -> PIXEL_YC
void f32_yc(PIXEL_YC *dst, const void *src, void *tmp, int w, int h, int max_w, int sw, int sh)
{
	int i,j;
	const float *fsrc = (const float*)src;
	
	for(i=0; i<h; ++i){
		for(j=0; j<w; ++j){
			dst->y  = (short)(4096.0f * fsrc[j*4 + 1]);
			dst->cb = (short)(4096.0f * fsrc[j*4 + 2]);
			dst->cr = (short)(4096.0f * fsrc[j*4 + 3]);
			dst++;
		}
		fsrc += sw*4;
		dst += (max_w - w);
	}

}

//ϊRs[ PIXEL_YC -> A16B16G16R16F SSEp
void yc_f16_sse(void *dst, const PIXEL_YC *src, void *tmp, int w, int h, int max_w, int sw, int sh)
{
	int i,j;
	__m64 m0;
	__m128 x0;
	float *ftmp = (float*)tmp;
	static const __m128 xr = { 0.0f, 1.0f/4096.0f, 1.0f/4096.0f, 1.0f/4096.0f };
	D3DXFLOAT16 *stmp = (D3DXFLOAT16*)dst;

	for(i=0; i<h; ++i){
		m0 = *(__m64*)src;
		for(j=0; j<w; ++j){
			m0 = _mm_slli_si64(m0, 16);
			x0 = _mm_cvtpi16_ps(m0);

			src++;
			m0 = *(__m64*)src;

			x0 = _mm_mul_ps(x0, xr);
			_mm_store_ps(ftmp + 4*j, x0);
		}
		D3DXFloat32To16Array(stmp, ftmp, w*4);

		src += (max_w - w);
		stmp += sw*4;
	}

	_mm_empty();
}

//ϊRs[ A16B16G16R16F -> PIXEL_YC SSEp
void f16_yc_sse(PIXEL_YC *dst, const void *src, void *tmp, int w, int h, int max_w, int sw, int sh)
{
	int i,j;
	__m64 m0;
	__m128 x0;
	float *ftmp = (float*)tmp;
	static const __m128 xc = { 0.0f, 4096.0f, 4096.0f, 4096.0f };
	const D3DXFLOAT16 *stmp = (const D3DXFLOAT16*)src;
	
	for(i=0; i<h; ++i){
		D3DXFloat16To32Array(ftmp, stmp, w*4);
		for(j=0; j<w; ++j){
			x0 = _mm_load_ps(ftmp + 4*j);
			x0 = _mm_mul_ps(x0, xc);

			m0 = _mm_cvtps_pi16(x0);
			m0 = _mm_srli_si64(m0, 16);
			*(__m64*)dst = m0;
			++dst;
		}
		stmp += sw*4;
		dst += (max_w - w);
	}

	_mm_empty();
}

//ϊRs[ PIXEL_YC -> A32B32G32R32F SSEp
void yc_f32_sse(void *dst, const PIXEL_YC *src, void *tmp, int w, int h, int max_w, int sw, int sh)
{
#if 1
	int i,j;
	__m64 m0;
	__m128 x0;
	__m128 *fdst = reinterpret_cast<__m128*>(dst);
	static const __m128 xr = { 0.0f, 1.0f/4096.0f, 1.0f/4096.0f, 1.0f/4096.0f };

	for(i=0; i<h; ++i){
		m0 = *(__m64*)src;
		for(j=0; j<w; ++j){
			m0 = _mm_slli_si64(m0, 16);
			x0 = _mm_cvtpi16_ps(m0);

			src++;
			m0 = *(__m64*)src;

			x0 = _mm_mul_ps(x0, xr);
			//_mm_store_ps((float*)(fdst + j), x0);
			_mm_storeu_ps((float*)(fdst + j), x0);
			//_mm_stream_ps((float*)(fdst + j), x0);
		}
		src += (max_w - w);
		fdst += sw;
	}
#else //ǂݍݎ̃ACg킹ǂ܂ʂȂ
	int i,j;
	__m64 m0,m1,m2,m3;
	__m128 x0,x1,x2,x3;
	__m64 *msrc;
	__m128 *fdst = (__m128*)dst;
	static const __m128 xr = { 0.0f, 1.0f/4096.0f, 1.0f/4096.0f, 1.0f/4096.0f };

	for(i=0; i<h; ++i){
		msrc = (__m64*)src;
		m0 = *msrc++;
		for(j=0; j<w; j+=4){
			m1 = *msrc++;

			m3 = _mm_slli_si64(m0, 16);
			x0 = _mm_cvtpi16_ps(m3);

			x0 = _mm_mul_ps(x0, xr);
			_mm_stream_ps((float*)(fdst + j), x0);

			m0 = _mm_srli_si64(m0, 32);
			m0 = _mm_unpacklo_pi32(m0, m1);
			x1 = _mm_cvtpi16_ps(m0);
			m2 = *msrc++;

			x1 = _mm_mul_ps(x1, xr);
			_mm_stream_ps((float*)(fdst + j + 1), x1);

			m1 = _mm_srli_si64(m1, 32);
			m1 = _mm_unpacklo_pi32(m1, m2);
			m1 = _mm_slli_si64(m1, 16);
			x2 = _mm_cvtpi16_ps(m1);
			m0 = *msrc++;
			
			x2 = _mm_mul_ps(x2, xr);
			_mm_stream_ps((float*)(fdst + j + 2), x2);
			
			x3 = _mm_cvtpi16_ps(m2);

			x3 = _mm_mul_ps(x3, xr);
			_mm_stream_ps((float*)(fdst + j + 3), x3);
		}
		src += max_w;
		fdst += sw;
	}
#endif
	_mm_empty();
}

//ϊRs[ A32B32G32R32F -> PIXEL_YC SSEp
void f32_yc_sse(PIXEL_YC *dst, const void *src, void *tmp, int w, int h, int max_w, int sw, int sh)
{
	int i,j;
	__m64 m0;
	__m128 x0;
	const __m128 *fsrc = (const __m128*)src;
	static const __m128 xc = { 0.0f, 4096.0f, 4096.0f, 4096.0f };
	
	for(i=0; i<h; ++i){
		for(j=0; j<w; ++j){
			x0 = _mm_load_ps((float*)(fsrc + j));
			//x0 = _mm_loadu_ps((float*)(fsrc + j));
			x0 = _mm_mul_ps(x0, xc);

			m0 = _mm_cvtps_pi16(x0);
			m0 = _mm_srli_si64(m0, 16);
			*(__m64*)dst = m0;
			++dst;
		}
		fsrc += sw;
		dst += (max_w - w);
	}

	_mm_empty();
}

//ϊRs[ PIXEL_YC -> D3DFMT_A16B16G16R16 MMXp
static void yc_i16_mmx(void *dst, const PIXEL_YC *src, void *tmp, int w, int h, int max_w, int sw, int sh)
{
	int i,j;
	__m64 m0;
	__m64 *mdst = (__m64*)dst;

	for(i=0; i<h; ++i){
		for(j=0; j<w; ++j){
			m0 = *(__m64*)(src + j);
			mdst[j] = _mm_slli_si64(m0, 16);
			//_mm_stream_pi(mdst + j, _mm_slli_si64(m0, 16));
		}
		src += max_w;
		mdst += sw;
	}

	_mm_empty();
}

//ϊRs[ D3DFMT_A16B16G16R16 -> PIXEL_YC MMXp
static void i16_yc_mmx(PIXEL_YC *dst, const void *src, void *tmp, int w, int h, int max_w, int sw, int sh)
{
	int i,j;
	__m64 m0;
	const __m64 *msrc = (const __m64*)src;
	
	for(i=0; i<h; ++i){
		for(j=0; j<w; ++j){
			m0 = _mm_srli_si64(msrc[j], 16);
			*(__m64*)dst++ = m0;
			//_mm_stream_pi((__m64*)dst++, m0);
		}
		msrc += sw;
		dst += (max_w - w);
	}

	_mm_empty();
}

//ϊRs[ D3DFMT_A2B10G10R10 -> PIXEL_YC MMXp
static void i10_yc_mmx(PIXEL_YC *dst, const void *src, void *tmp, int w, int h, int max_w, int sw, int sh)
{
	int i,j;
	const UINT *dsrc = (const UINT*)src;
	int d,c;
	const int cf = 4096 / 512;
	const int cg = 4096 / 2048;
	
	//TODO:MMXĂ܂!!
	for(i=0; i<h; ++i){
		for(j=0; j<w; ++j){
			d = dsrc[j];

			c = (d & 0xfff00000) >> 20;
			c -= 1024;
			dst->y  = (c * cg);
			c = (d & 0xffc00) >> 10;
			c = c>511 ? c-1024 : c;
			dst->cb = (c * cf);
			c = (d & 0x3ff);
			c = c>511 ? c-1024 : c;
			dst->cr = (c * cf);

			dst++;
		}
		dsrc += sw;
		dst += (max_w - w);
	}
}

//SSEɑΉĂ邩`FbN
static bool isSSEavailable(void)
{
	long r_ecx,r_edx;

	// CPUIDɂ`FbN
	__asm {
		mov eax, 1
		cpuid
		mov r_edx, edx
		mov r_ecx, ecx
	}

	// MMX  0x0800000 edx
	// SSE  0x2000000 edx
	// SSE2 0x4000000 edx
	// SSE3 0x0000001 ecx
	r_edx &= 0x6800000;
	r_ecx &= 0x0000001;

	if(r_edx < 0x1000000)
	return false;

	// OST|[gĂ邩
	__try {
		__asm xorps xmm0, xmm0
	}

	__except(EXCEPTION_EXECUTE_HANDLER){
		if (_exception_code()==STATUS_ILLEGAL_INSTRUCTION)
			/* SSE not supported */
			return false;
	}
	/* SSE are supported by OS */
	return true;
}

//ɃI[vĂD3DC^[tF[X擾
// ۂ͕KvȂHH
typedef BOOL (__stdcall * GETD3D9IF)(IDirect3D9 **, IDirect3DDevice9 **);
bool searchD3D(FILTER *fp, IDirect3D9* &pD3D, IDirect3DDevice9* &pd3dDevice)
{
	//tB^𓾂
	SYS_INFO si;
	if(FALSE == fp->exfunc->get_sys_info(NULL, &si)){
		return false;
	}
	
	//̃vOCœ̊֐T|[gĂ̂
	//C^[tF[X擾
	IDirect3D9 *pTmpD3D = NULL;
	IDirect3DDevice9 *pTmpd3dDevice = NULL;
	FILTER *pTmpF = NULL;
	GETD3D9IF pFunc = NULL;
	for(int i=0; i<si.filter_n; ++i){
		pTmpF = reinterpret_cast<FILTER*>(fp->exfunc->get_filterp(i));
		if(pTmpF){
			pFunc = reinterpret_cast<GETD3D9IF>
				(GetProcAddress(pTmpF->dll_hinst, "GetD3D9Interface"));
			if(pFunc){
				if(TRUE == pFunc(&pTmpD3D, &pTmpd3dDevice)){
					//C^[tF[X̎擾
					break;
				}else{
					SAFE_RELEASE(pTmpD3D);
					SAFE_RELEASE(pTmpd3dDevice);
				}
			}
		}
	}

	//C^[tF[Xꂽ`FbN
	if(pTmpD3D && pTmpd3dDevice){
		pD3D = pTmpD3D;
		pd3dDevice = pTmpd3dDevice;
		return true;
	}else{
		return false;
	}
}
